local MAJOR = "StarLibDriverQTip-1.0"
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibDriverQTip = LibStub:NewLibrary(MAJOR, MINOR)
if not LibDriverQTip then return end
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibCore = LibStub("StarLibCore-1.0", true)
assert(LibCore, MAJOR .. " requires StarLibCore-1.0")
local LCDText = LibStub("StarLibLCDText-1.0", true)
assert(LCDText, MAJOR .. " requires StarLibLCDText-1.0")
local LibQTip = LibStub('LibQTip-1.0', true)
assert(LibQTip, MAJOR .. " requires LibQTip-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LibDriverQTip.__index then
	LibDriverQTip.__index = LibDriverQTip
end

-- @name StarLibDriverQTip:New
-- @usage StarLibDriverQTip:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor App
-- @param name A name for this LCD
-- @param config The configuration
-- @param errorLevel Error verbosity
-- @return A new StarLibDriverQTip object

function LibDriverQTip:New(visitor, name, config, errorLevel)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.buffer = {}
		obj.environment = {}
	end

	setmetatable(obj, self)
	
	obj.core = LibCore:New(obj, obj.environment, name, config, "text", errorLevel)
	obj.error = LibError:New(MAJOR, errorLevel)
	local rows, cols = config[name].rows, config[name].cols
	obj.rows, obj.cols, obj.yres, obj.xres, obj.layers = config[name].rows or 4, config[name].cols or 20, config[name].yres or 8, config[name].xres or 7, config[name].layers or 3
	obj.lcd = LCDText:New(obj.core, obj.rows, obj.cols, obj.yres, obj.xres, obj.layers, errorLevel, obj.Blit, obj)
	obj.core.lcd = obj.lcd -- You must provide a LibCore object with an LCD object.
	obj.tooltip = LibQTip:Acquire("StarLibDriverQTip", obj.cols, "LEFT", "CENTER", "RIGHT")
	obj.tooltip:SmartAnchorTo(UIParent)
	
	-- New font looking like GameTooltipText but red with height 15
	local font = CreateFont("LibDriverQTip")
	font:SetFont(GameTooltipText:GetFont(), 12)
	font:SetTextColor(1,1,1)
	obj.tooltip:SetFont(font)
	
	local tbl = {}
	for i = 1, obj.cols do
		tinsert(tbl, ' ')
	end
	for i = 0, obj.rows - 1 do
		obj.tooltip:AddLine(unpack(tbl))
	end
		
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.core:CFGSetup()
	obj.core:BuildLayouts()
	
	return obj	
	
end

-- Delete an object
-- @name StarLibDriverQTip:Del
-- @usage StarLibDriverQTip:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LibDriverQTip:Del(lcd)
	if not lcd then
		lcd = self
	end
	pool[lcd] = true
	for i, v in ipairs(objects) do
		if v == lcd then
			objectsDict[lcd] = nil
			tremove(objects, i)
		end
	end
	lcd.tooltip:Hide()
	LibQTip:Release(lcd.tooltip)
	lcd.core:Del()
	lcd.error:Del()
end

-- Show the tooltip
-- @name StarLibDriverQTip:Show
-- @usage StarLibDriverQTip:Show()
-- @return Nothing
function LibDriverQTip:Show()
	self.tooltip:Show()
	self.core:Start()
end

-- Hide the tooltip
-- @name StarLibDriverQTip:Hide
-- @usage StarLibDriverQTip:Hide()
-- @return Nothing
function LibDriverQTip:Hide()
	self.tooltip:Hide()
	self.core:Stop()
end

function LibDriverQTip:Blit(obj, r, c, buffer, len)
	for i = c, c + len - 1 do
		obj.tooltip:SetCell(r + 1, i + 1, (buffer[i - c] or '?'):sub(1, 1))
	end
end