local MAJOR = "StarLibWidgetText-1.0" 
local MINOR = 1

assert(LibStub, MAJOR.." requires LibStub") 
local WidgetText = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetText then return end

local WidgetPlugin = LibStub:NewLibrary("StarLibWidgetTextPlugin-1.0", 1)

local LibProperty = LibStub("StarLibProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibWidget = LibStub("StarLibWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires StarLibWidget-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")
local LibRGBA = LibStub("StarLibRGBA-1.0", true)
assert(LibRGBA, MAJOR .. " requires StarLibRGBA-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")

local PINGPONGWAIT = 2

WidgetText.ALIGN_LEFT, WidgetText.ALIGN_CENTER, WidgetText.ALIGN_RIGHT, WidgetText.ALIGN_MARQUEE, WidgetText.ALIGN_AUTOMATIC, WidgetText.ALIGN_PINGPONG = 1, 2, 3, 4, 5, 6
local alignmentList = {"Left", "Center", "Right", "Marquee", "Automatic", "Pingpong"}
local alignmentDict = {Left = 1, Center = 2, Right = 3, Marquee = 4, Automatic = 5, Pingpong = 6}

WidgetText.alignmentList = alignmentList
WidgetText.alignmentDict = alignmentDict

WidgetText.SCROLL_RIGHT, WidgetText.SCROLL_LEFT = 1, 2
local directionList = {"Right", "Left"}
local directionDict = {Right = 1, Left = 2}

WidgetText.directionList = directionList
WidgetText.directionDict = directionDict

local map = {value = "Value", prefix = "Prefix", postfix = "Postfix", precision = "Precision", align = "Alignment", update = "Update", speed = "Scroll Speed", direction ="Direction", cols = "Columns"}

local widgetType = {text = true, rc = true}

WidgetPlugin.New = function(environment)
	environment.ALIGN_LEFT, environment.ALIGN_CENTER, environment.ALIGN_RIGHT, environment.ALIGN_MARQUEE, environment.ALIGN_AUTOMATIC, environment.ALIGN_PINGPONG = 1, 2, 3, 4, 5, 6
	environment.SCROLL_RIGHT, environment.SCROLL_LEFT = 1, 2
end

local textUpdate, scrollUpdate
local pool = setmetatable({}, {__mode = "k"})

if not WidgetText.__index then
	WidgetText.__index = WidgetText
end

--[[
local function resizeText(str, cols)
	if strlen(str) < cols then
		for i = strlen(str) + 1, cols do
			str = str .. " "
		end
	end
	return str:sub(1, cols)
end

local function insertChar(text1, pos1, text2, pos2)
	local char = text2:sub(pos2, pos2)
	
	local left = text1:sub(1, pos1 - 1)
	local right = text1:sub(pos1 + 1)
	
	return left .. char .. right
end
]]

WidgetText.defaults = {
	value = '',
	prefix = '',
	postfix = '',
	precision = 0xbabe,
	align = WidgetText.ALIGN_LEFT,
	update = 0,
	speed = 0,
	repeating = true,
	direction = WidgetText.SCROLL_RIGHT,
	cols = 40,
	color = {r=0xff, g=0xff, b=0xff, a=0xff}
}

-- Create a new StarLibWidgetText object
-- @name WidgetText.New
-- @usage WidgetText:New(visitor, name, config, row, col, layer, fontString, env, errorLevel)
-- @param visitor An StarLibCore-1.0 object
-- @param config This widget's settings
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param config The settings. You can provide these fields: string, prefix, postfix, precision, align, update, scroll, speed, direction, cols, offset, str, update, speed. Check the source for further details. Some fields have defaults.
-- @param self.errorLevel The self.errorLevel for this object
-- @param callback Your draw function. The widget is passed as first parameter and data is passed as 2nd.
-- @param data Some data to pass to the draw function
-- @return A new StarLibWidgetText object

function WidgetText:New(visitor, name, config, row, col, layer, errorLevel, callback, data) 
	assert(name, "WidgetText requires a name.")
	assert(config, "Please provide the marquee with a config")
	assert(config.value, name .. ": Please provide the marquee with a string")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.options = {}
	end
		
	setmetatable(obj, self)
	
	
	obj.widget = LibWidget:New(obj, visitor, name, config, row, col, layer, widgetType, env, errorLevel)
	obj.error = LibError:New(MAJOR .. " : " .. name, errorLevel)
	obj.color = LibRGBA:New()
	obj.color_valid = obj.widget:WidgetColor(config, "colorHex", obj.color);	
	obj.color.script = config.color
	obj.visitor = visitor
	obj.errorLevel = errorLevel or 3
	obj.value = LibProperty:New(visitor, name .. " string", config.value, "", errorLevel) -- text of marquee
	obj.prefix = LibProperty:New(visitor, name .. " prefix", config.prefix, "", errorLevel) -- label on the left side
	obj.postfix = LibProperty:New(visitor, name .. " postfix", config.postfix, "", errorLevel) -- label on right side
	obj.precision = config.precision or self.defaults.precision -- number of digits after the decimal point
	obj.align = config.align or self.defaults.align -- alignment: left, center, right, marquee, automatic, pingpong
	obj.update = config.update or self.defaults.update -- update interval
	obj.repeating = config.repeating or self.defaults.repeating -- Whether to use repeating timers
	obj.speed = config.speed or self.defaults.speed -- marquee scrolling speed
	obj.direction = config.direction or self.defaults.direction -- marquee direction	
	obj.cols = config.cols or self.defaults.cols -- number of colums in marquee
	obj.bold = config.bold
	obj.offset = 0 -- increment by pixel
	obj.string = "" -- formatted value
	obj.callback = callback
	obj.data = data
	obj.dontRtrim = config.dontRtrim
	if obj.direction == self.SCROLL_LEFT then
		obj.scroll = obj.cols -- marquee starting point
	else
		obj.scroll = 0
	end
	if obj.update > WidgetText.defaults.update then
		obj.timer = LibTimer:New("WidgetText.timer " .. obj.widget.name, obj.update, obj.repeating, textUpdate, obj)
	end
	
	if obj.speed > WidgetText.defaults.speed then
		obj.textTimer = LibTimer:New("WidgetText.textTimer " .. obj.widget.name, obj.speed, obj.repeating, textScroll, obj)
	end
		
	return obj	
end

-- Delete a StarLibWidgetText object
-- @name StarLibWidgetText.Del
-- @usage StarLibWidgetText:Del([marq])
-- @param marq A StarLibWidgetText object
function WidgetText:Del(marq)
	if not marq then
		marq = self
	end
	if marq.Stop then
		marq:Stop()
	end	
	if marq.widget then
		marq.widget:Del()
		marq.widget = nil
	end
	if marq.value then
		marq.value:Del()
		marq.value = nil
	end
	if marq.prefix then
		marq.prefix:Del()
		marq.prefix = nil
	end
	if marq.postfix then
		marq.postfix:Del()
		marq.postfix = nil
	end
	if marq.timer then
		marq.timer:Del()
		marq.timer = nil
	end
	if marq.textTimer then
		marq.textTimer:Del()
		marq.textTimer = nil
	end
	if marq._dst then
		marq._dst:Del()
		marq._src:Del()
		marq._dst = nil
		marq._src = nil
	end
	if marq.error then
		marq.error:Del()
		marq.error = nil
	end
	if marq.color then
		marq.color:Del()
		marq.color = nil
	end
	pool[marq] = true
end

-- Start a StarLibWidgetText
-- @name StarLibWidgetText.Start
-- @usage object:Start()
-- @return Nothing
function WidgetText:Start()
	self.error:Print("WidgetText:Start", 3)
	self:Update()
	if self.timer then
		self.timer:Start()
	end
	if self.textTimer then
		self.textTimer:Start()
	end
end

-- Stop a StarLibWidgetText
-- @name StarLibWidgetText.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetText:Stop()
	self.error:Print("WidgetText:Stop", 3)
	if self.timer then 
		self.timer:Stop()
	end
	if self.textTimer then
		self.textTimer:Stop()
	end
end

-- Update data
-- @name StarLibWidgetText.Update
-- @name StarLibWidgetText.Update
-- @usage StarLibWidgetText:Update()
-- @return True if we updated.
function WidgetText:Update()
	textUpdate(self)
	if self.speed > 0 then
		textScroll(self)
	end
end

-- Print text
-- @name StarLibWidgetText.Draw
-- @param text The text to print. Feel free to hook this.
-- @return Nothing
function WidgetText:Draw()
	if type(self.callback) == "function" then
		self:callback(self.data)
	end
end

local function rtrim(text)
	local pos = 0
	
	for i = 1, strlen(text) do
		if text:sub(i, i) ~= ' ' then
			pos = i
		end
	end
	
	if pos == 0 then
		return strtrim(text)
	else
		return text:sub(1, pos) .. strtrim(text:sub(pos + 1))
	end
end


function textScroll(self)
	if not self.prefix or not self.postfix then return end

	self.count = (self.count or 0) + 1
	self._pre = self.prefix:P2S()
	self._post = self.postfix:P2S()

	self._str = self.string .. ' '
	
	self._num, self._len, self._width, self._pad = 0, strlen(self._str), self.cols - strlen(self._pre) - strlen(self._post), 0
	
	self._srcPtr, self._dstPtr = 1, 1
	
	if not self._src then
		self._src = LibBuffer:New(MAJOR .. ": " .. self.widget.name .. "._src", 0, self.errorLevel)
		self._dst = LibBuffer:New(MAJOR .. ": " .. self.widget.name .. "._dst ", 0, self.errorLevel)
	end
	
	self._src:Clear()
	self._dst:Clear()
	
    if self._width < 0 then
        self._width = 0
	end
		
	if self.align == self.ALIGN_LEFT then
		self._pad = 0
	elseif self.align == self.ALIGN_CENTER then
		self._pad = (self._width - self._len) / 2
		if self._pad < 0 then
			self._pad = 0
		end
	elseif self.align == self.ALIGN_RIGHT then
		self._pad = self._width - self._len
		if self._pad < 0 then
			self._pad = 0
		end
	elseif self.align == self.ALIGN_AUTOMATIC then
		if self._len <= self._width then
			self._pad = 0
		end
	elseif self.align == self.ALIGN_MARQUEE then
		self._pad = self._width - self.scroll
		if self.direction == self.SCROLL_LEFT then
			self._pad = self._pad + strlen(self._str)
			self.scroll = self.scroll + 1
			if self.scroll >= self._width then
				self.scroll = 0
			end
			
		else
			self.scroll = self.scroll - 1
			if self.scroll < 0 then 
				self.scroll = self.cols 
			end
		end
	elseif self.align == self.ALIGN_PINGPONG then
		if self._len <= self._width then
			self._pad = (self._width - self._len) / 2
		else
			if self.direction == self.SCROLL_RIGHT then
				self.scroll = self.scroll + 1
			else
				self.scroll = self.scroll - 1
			end
			
			self._pad = 0 - self.scroll;
			
			if self._pad < 0 - (self._len - self._width) then
				if self.delay < 1 then
					self.direction = self.SCROLL_LEFT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll - PINGPONGWAIT
				end
				self._pad = 0 - (self._len - self._width)
			elseif self._pad > 0 then
				if self.delay < 1 then
					self.direction = self.SCROLL_RIGHT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll + PINGPOINGWAIT
				end
				self._pad = 0
			else
				self._pad = 0
			end
		end
	else
		self._pad = 0
		self.error:Print("No alignment specified")
	end

	self._dst:Resize(self.cols) --LibUtils:resizeText(self._dst, self.cols)
	
    self._dstPtr = 0;
    self._num = 0;	
	
    -- /* process prefix */
    self._src:FromString(self._pre)
    while (self._num < self.cols) do
        if (self._srcPtr >= self._src:Size()) then
            break
		end
		if self._dstPtr >= self.cols then
			break
		end
		self._dst:Replace(self._dstPtr, self._src.buffer[self._srcPtr]) --LibUtils:replaceText(self._dst, self._dstPtr, self._src, self._srcPtr)
		self._dstPtr = self._dstPtr + 1
		self._srcPtr = self._srcPtr + 1
        self._num = self._num + 1
    end

    self._src:FromString(self._str)
    self._srcPtr = 0;
	
	self._offset = self._pad
	
    if(self._offset < 0) then
        self._offset = 0;
	end
	
    -- /* wrap around on the beginning */
    while (self._pad > 0 and self._num <= self.cols) do
		if self._dstPtr >= self.cols then
			break
		end
        if(self.align == self.ALIGN_MARQUEE) then
			self._dst:Replace(self._dstPtr, self._src.buffer[(self._src:Size() - self._offset) + self._srcPtr])-- LibUtils:replaceText(self._dst, self._dstPtr, self._src, (strlen(self._src) - self._offset) + self._srcPtr)
            --dst[dstPtr] = src[(strlen(src) - offset) + srcPtr];
			self._srcPtr = self._srcPtr + 1			
        else
            self._dst:Replace(self._dstPtr, " ")-- = LibUtils:replaceText(self._dst, self._dstPtr, " ", 1)
		end
		self._dstPtr = self._dstPtr + 1
        self._num = self._num + 1
        self._pad = self._pad - 1
    end

	--[[
    --/* skip src chars (marquee) */
	self._tmp = LibBuffer:New(MAJOR .. "._tmp", 0, self.errorLevel)
	self._tmp:Copy(self._src)
	
	--self._tmp = self._src
	
    while (self._pad < 0 and self._tmp:AsString() ~= "") do
		local tmp = self._tmp:Sub(1, self._tmp:Size())
        self._src:Copy(tmp)
        self._tmp:Copy(self._src)
		self._tmp:Del()
        self._pad = self._pad + 1
    end
		
	self._tmp:Del()
	]]
	
    --/* copy content */
    while (self._num < self.cols and self.align ~= self.ALIGN_MARQUEE and self.align ~= self.ALIGN_PINGPONG) do
		if self._dstPtr >= self.cols then
			break
		end	
        if (self._srcPtr >= self._src:Size()) then
            break;
		end
		self._dst:Replace(self._dstPtr, self._src.buffer[self._srcPtr])-- = LibUtils:replaceText(self._dst, self._dstPtr, self._src, self._srcPtr)
        --dst[dstPtr] = src[srcPtr];
		self._dstPtr = self._dstPtr + 1
		self._srcPtr = self._srcPtr + 1
        self._num = self._num + 1
    end
	
    -- wrap around on end 
    self._src:FromString(self._post)
    self._len = self._src:Size()
    self._srcPtr = 0;
	
    if(self._num < self.cols - self._len and self.align == ALIGN_MARQUEE) then
		self._dst:Replace(self._dstPtr, "*") -- = LibUtils:replaceText(self._dst, self._dstPtr, "*", 1)
        --dst[dstPtr] = '*';
		self._dstPtr = self._dstPtr + 1
        self._num = self._num + 1
    end

    while (self._num < self.cols - self._len) do
		if self._dstPtr >= self.cols then
			break
		end
		if self._srcPtr >= self._src:Size() then
			break
		end
        if(self.align == self.ALIGN_MARQUEE) then
			self._dst:Replace(self._dstPtr, self._src.buffer[self._srcPtr]) -- = LibUtils:replaceText(self._dst, self._dstPtr, self._src, self._srcPtr)
            --dst[dstPtr] = str[srcPtr];
			self._dstPtr = self._dstPtr + 1
			self._srcPtr = self._srcPtr + 1
        else
			self._dst:Replace(self._dstPtr, " ") -- LibUtils:replaceText(self._dst, self._dstPtr, ' ', 1)
            --dst[dstPtr] = ' ';
			self._dstPtr = self._dstPtr + 1
		end
        self._num = self._num + 1;
    end

    self._srcPtr = 0;

    --/* process postfix */
    while (self._num < self.cols) do
		if self._dstPtr >= self.cols then
			break
		end	
        if (self._srcPtr >= self._src:Size()) then
            break;
		end
		self._dst:Replace(self._dstPtr, self._src.buffer[self._srcPtr]) -- = LibUtils:replaceText(self._dst, self._dstPtr, self._src, self._srcPtr)
        --dst[dstPtr] = src[srcPtr];
		self._dstPtr = self._dstPtr + 1
		self._srcPtr = self._srcPtr + 1
        self._num = self._num + 1
    end
	
	if self.dontRtrim then
		self.buffer = self._dst:AsString()
	else
		self.buffer = rtrim(self._dst:AsString())
	end
		
	if strlen(self.buffer) ~= self.cols and self.DontRtrim then
		StarTip:Print("'" .. self.buffer .. "'")
	end
	
	if self.buffer ~= self.oldBuffer then
		self:Draw()
		self.oldBuffer = self.buffer
	end
end


function textUpdate(self)
	self._update = 0
	
	assert(self.value, "WidgetText has no value")
	if not self.prefix then return self.error:Print("WidgetText needs a prefix")end
	if not self.postfix then return self.error:Print("WidgetText needs a postfix") end
	
	self._update = self._update + self.prefix:Eval()
	self._update = self._update + self.postfix:Eval()
	
	self.value:Eval()
	
    -- /* str or number? */
    if (self.precision == 0xBABE) then
        self._str = self.value:P2S();
    else
        local number = self.value:P2N();
        local width = self.cols - strlen(self.prefix:P2S()) - strlen(self.postfix:P2S());
        local precision = self.precision;
        --[[/* print zero bytes so we can specify NULL as target  */
        /* and get the length of the resulting str */]]
		local text = ("%.*f"):format(precision, number)
		local size = strlen(text)
        --/* number does not fit into field width: try to reduce precision */
        if (width < 0) then
            width = 0;
		end
        if (size > width and precision > 0) then
            local delta = size - width;
            if (delta > precision) then
                delta = precision;
			end
            precision = precision - delta;
            size = size - delta;
            --/* zero precision: omit decimal point, too */
            if (precision == 0) then
                size = size - 1
			end
        end
        ---/* number still doesn't fit: display '*****'  */
        if (size > width) then
            str.resize(width);
            for i = 0, width do
                str[i] = '*';
			end
        else
            str = text
        end
    end

    if self._str == "" or self._str ~= self.string then 
        self._update = self._update + 1;
        self.string = self._str;
    end

    --/* something has changed and should be updated */
    if (self._update > 0) then
        -- /* Init pingpong scroller. start scrolling left (wrong way) to get a delay */
        if (self.align == ALIGN_PINGPONG) then
            self.direction = 0;
            self.delay = PINGPONGWAIT;
        end
		--[[
        /* if there's a marquee scroller active, it has its own */
        /* update callback timer, so we do nothing here; otherwise */
        /* we simply call this scroll callback directly */
		]]
        if (self.align ~= ALIGN_MARQUEE and self.align ~= ALIGN_AUTOMATIC and self.align ~= ALIGN_PINGPONG) then
            textScroll(self)
			return false
        end

    end

	return true
end

function WidgetText:GetOptions(db)
		local defaults = WidgetText.defaults
		local options = {
			marquee = {
				name = "Enabled",
				type = "toggle",
				get = function() return db.marquee end,
				set = function(info, v) db.marquee = v end,
				order = 1
			},
			value = {
				name = "Value",
				type = "input",
				multiline = true,
				get = function()
					return db.value or defaults.value
				end,
				set = function(info, v)
					db.value = v
				end,
				order = 2
			},
			prefix = {
				name = "Prefix",
				type = "input",
				multiline = true,
				get = function()
					return db.prefix or defaults.prefix
				end,
				set = function(info, v)
					db.prefix = v
				end,
				order = 3
			},
			postfix = {
				name = "Postfix",
				type = "input",
				multiline = true,
				get = function()
					return db.postfix or defaults.postfix
				end,
				set = function(info, v)
					db.postfix = v
				end,
				order = 4
			},
			precision = {
				name = "Precision",
				type = "input",
				pattern = "%d",
				get = function()
					return db.precision or defaults.precision
				end,
				set = function(info, v)
					db.precision = v
					self.precision = v
				end,
				order = 5
			},
			align = {
				name = "Alignment",
				type = "select",
				values = alignmentText,
				get = function()
					return db.align or defaults.align
				end,
				set = function(info, v)
					db.align = v
					self.align = v
				end,
				order = 6
			},
			update = {
				name = "Update",
				type = "input",
				pattern = "%d",
				get = function() 
					return db.update or defaults.update
				end,
				set = function(info, v)
					db.update = v
				end,
				order = 7
			},
			speed = {
				name = "Scroll Speed",
				type = "input",
				pattern = "%d",
				get = function()
					return db.speed or defaults.speed
				end,
				set = function(info, v)
					db.speed = v
				end,
				order = 8
			},
			direction = {
				name = "Direction",
				type = "select",
				values = directions,
				get = function()
					return db.direction or defaults.direction
				end,
				set = function(info, v)
					db.direction = v
				end,
				order = 9
			},
			cols = {
				name = "Columns",
				type = "input",
				pattern = "%d",
				get = function()
					return db.cols or defaults.cols
				end,
				set = function(info, v)
					db.cols = v
				end,
				order = 10
			},
			color = {
				name = "Color",
				type = "color",
				get = function()
					if not db.color then db.color = {} end
					return (db.color.r or defaults.color.r) / 255, (db.color.g or defaults.color.g) / 255, (db.color.b or defaults.color.b) / 255, (db.color.a or defaults.color.a) / 255
				end,
				set = function(info, r, g, b, a)
					if not db.color then db.color = {} end
					db.color.r = r * 255
					db.color.g = g * 255
					db.color.b = b * 255
					db.color.a = a * 255
				end,
				order = 11
			},
			bold = {
				name = "Bold",
				type = "toggle",
				get = function()
					return db.bold
				end,
				set = function(info, val)
					db.bold = val
				end,
				order = 12
			}
		}
	return options
end
